from django.conf import settings


PAYPAL_MODE = getattr(settings, "PAYPAL_MODE", None)
PAYPAL_CLIENT_ID = getattr(settings, "PAYPAL_CLIENT_ID", None)
PAYPAL_CLIENT_SECRET = getattr(settings, "PAYPAL_CLIENT_SECRET", None)

PAYPAL_LIVE_MODE = PAYPAL_MODE == "live"

PAYPAL_WEBHOOK_ID = getattr(settings, "PAYPAL_WEBHOOK_ID", "")

PAYPAL_SETTINGS = {
	"mode": PAYPAL_MODE,
	"client_id": PAYPAL_CLIENT_ID,
	"client_secret": PAYPAL_CLIENT_SECRET,
}

if PAYPAL_CLIENT_ID and PAYPAL_CLIENT_SECRET:
	from paypalrestsdk import configure
	configure(PAYPAL_SETTINGS)



from django.core.management import BaseCommand
from paypalrestsdk import payments as paypal_models

from djpaypal import models


class Command(BaseCommand):
	help = "Syncs plan data from upstream Paypal"

	def handle(self, *args, **options):
		for status in ("created", "active"):
			all_plans = paypal_models.BillingPlan.all({"status": status})
			if all_plans.plans is None:
				continue
			models.BillingPlan.objects.sync_data(all_plans.plans)



from django.db import models
from django.urls import reverse

from ..settings import PAYPAL_LIVE_MODE


class PaypalObjectManager(models.Manager):
	def sync_data(self, paypal_data, fetch=True):
		ret = []
		for obj in paypal_data:
			if fetch:
				api_full_data = self.model.paypal_model.find(obj["id"])
			else:
				api_full_data = obj
			db_obj, _ = self.model.get_or_update_from_api_data(api_full_data)
			ret.append(db_obj)
		return ret


class PaypalObject(models.Model):
	class Meta:
		abstract = True

	id = models.CharField(max_length=128, primary_key=True, editable=False, serialize=True)
	livemode = models.BooleanField()

	djpaypal_created = models.DateTimeField(auto_now_add=True)
	djpaypal_updated = models.DateTimeField(auto_now=True)

	objects = PaypalObjectManager()

	id_field_name = "id"
	dashboard_url_template = ""

	@staticmethod
	def sdk_object_as_dict(obj):
		if isinstance(obj, dict):
			return obj.copy()
		# Paypal SDK object
		return obj.to_dict()

	@classmethod
	def clean_api_data(cls, data):
		cleaned_data = cls.sdk_object_as_dict(data)

		# Delete links (only useful in the API itself)
		if "links" in cleaned_data:
			del cleaned_data["links"]

		# Extract the ID to return it separately
		id = cleaned_data.pop(cls.id_field_name)

		# Set the livemode
		cleaned_data["livemode"] = PAYPAL_LIVE_MODE

		return id, cleaned_data, {}

	@classmethod
	def get_or_update_from_api_data(cls, data, always_sync=False):
		id, cleaned_data, m2ms = cls.clean_api_data(data)
		db_obj, created = cls.objects.get_or_create(**{
			cls.id_field_name: id,
			"defaults": cleaned_data,
		})
		if always_sync or not created:
			db_obj.sync_data(cleaned_data)
			db_obj.save()

		for field, objs in m2ms.items():
			for obj in objs:
				getattr(db_obj, field).add(obj)

		return db_obj, created

	@classmethod
	def find_and_sync(cls, id):
		obj = cls.paypal_model.find(id)
		db_obj, created = cls.get_or_update_from_api_data(obj, always_sync=True)
		return db_obj

	def __str__(self):
		if hasattr(self, "name") and self.name:
			return self.name
		return self.id

	@property
	def admin_url(self):
		return reverse(
			"admin:{app_label}_{model_name}_change".format(
				app_label=self._meta.app_label, model_name=self._meta.model_name
			),
			args=(self.id, )
		)

	@property
	def dashboard_url(self):
		if self.livemode:
			paypal_url = "https://www.paypal.com"
		else:
			paypal_url = "https://www.sandbox.paypal.com"

		return self.dashboard_url_template.format(
			paypal=paypal_url, webscr=paypal_url + "/cgi-bin/webscr",
			id=getattr(self, self.id_field_name)
		)

	def _sync_data_field(self, k, v):
		if k == "links":
			return False

		if getattr(self, k) != v:
			setattr(self, k, v)
			return True

		return False

	def find_paypal_object(self):
		return self.paypal_model.find(self.id)

	def sync_data(self, obj):
		obj = self.sdk_object_as_dict(obj)
		updated = False
		for k, v in obj.items():
			if self._sync_data_field(k, v):
				updated = True

		if updated:
			self.save()



# Generated by Django 2.0.3 on 2018-03-20 07:02

from django.db import migrations, models


class Migration(migrations.Migration):
	dependencies = [
		("djpaypal", "0002_payer_time_created"),
	]

	operations = [
		migrations.AddField(
			model_name="dispute",
			name="dispute_channel",
			field=models.CharField(blank=True, editable=False, max_length=8),
		),
		migrations.AddField(
			model_name="dispute",
			name="dispute_life_cycle_stage",
			field=models.CharField(blank=True, editable=False, max_length=15),
		),
	]



# Generated by Django 3.0.1 on 2020-01-30 11:19

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djpaypal', '0012_billing_agreement_state'),
    ]

    operations = [
        migrations.AddField(
            model_name='billingagreement',
            name='plan_model',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='djpaypal.BillingPlan'),
        ),
        migrations.RunSQL("""
            UPDATE djpaypal_billingagreement db
            SET plan_model_id = dp."data"->'plan'->>'id'
            FROM djpaypal_preparedbillingagreement dp
            WHERE dp.executed_agreement_id = db.id;
        """)
    ]



# Generated by Django 3.0.1 on 2019-12-30 11:58

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djpaypal', '0011_disputes'),
    ]

    operations = [
        migrations.RunSQL("UPDATE djpaypal_billingagreement SET state = 'Cancelled' WHERE state = 'Canceled';")
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['djpaypal',
 'djpaypal.management',
 'djpaypal.management.commands',
 'djpaypal.migrations',
 'djpaypal.models']

package_data = \
{'': ['*'], 'djpaypal': ['templates/djpaypal/admin/*']}

install_requires = \
['django>=2.2', 'paypalrestsdk>=1.13.1', 'python-dateutil>=2.6.1']

setup_kwargs = {
    'name': 'dj-paypal',
    'version': '0.15.0',
    'description': 'Django / Paypal integration (based on dj-stripe)',
    'long_description': '# dj-paypal\n\n[![Build Status](https://travis-ci.com/HearthSim/dj-paypal.svg?branch=master)](https://travis-ci.com/HearthSim/dj-paypal)\n[![PyPI](https://img.shields.io/pypi/v/dj-paypal.svg)](https://pypi.org/project/dj-paypal/)\n\n\nA Paypal integration for Django, inspired by [dj-stripe](https://github.com/dj-stripe/dj-stripe).\n\nCurrently only supports subscriptions.\n\n\n## Requirements\n\n- Python 3.6+\n- Django 2.0+\n- Postgres 9.6+ (Non-postgres engines not supported)\n\n\n## Installation\n\n1. Install dj-paypal with `pip install dj-paypal`\n2. Add `djpaypal` to django `INSTALLED_APPS` setting\n3. Get a client ID and client secret from paypal and add them to the settings\n   `PAYPAL_CLIENT_ID` and `PAYPAL_CLIENT_SECRET`\n4. Set `PAYPAL_MODE = "sandbox"` (or `"live"`) in the settings\n5. Install your Billing Plans (see below)\n\n\n### Setting up billing plans\n\n#### Download already-created billing plans from Paypal\n\nRun `manage.py djpaypal_download_plans` to sync all plans already created upstream\ninto the local database.\n\nThis will create `djpaypal.models.BillingPlan` objects, which can be listed from\nthe Django admin.\n\n\n#### Creating new Paypal billing plans\n\nThe `manage.py djpaypal_upload_plans` management command creates billing plans using\nthe Paypal API. An extra `PAYPAL_PLANS` setting must be set, which will contain a dict\nof Paypal billing plans to create.\n\nSee `example_settings.py` for an example of plans to create.\n\n\n## Webhooks\n\nThe `djpaypal.views.ProcessWebhookView` view should be hooked up to an URL endpoint\nwhich you then set up in Paypal as a webhook endpoint (https://developer.paypal.com).\n\nIn order to verify webhooks being transmitted to your app, dj-paypal needs to know the\nID of the webhook that is expected at that URL. Set it in the setting `PAYPAL_WEBHOOK_ID`.\n\n\n## Sandbox vs. Live\n\nAll models have a `livemode` boolean attribute. That attribute is set to `True` if created\nin Live (production) mode, `False` otherwise (sandbox mode).\nSandbox and Live data can co-exist without issues. Once you are done testing in Sandbox\nmode, use the `manage.py djpaypal_delete_all_test_data` management command to (locally)\nclear all the test data. This command has no impact on the upstream data.\n\n\n## Data considerations\n\nMost of the models defined in dj-paypal are copies of the upstream Paypal model data.\nDeleting or editing objects (be it from the admin or in the database) does not actually\nchange any of the upstream Paypal data.\n\n\n## License and Sponsorship\n\nThis project was designed and developed by [HearthSim](https://hearthsim.net). It is\nlicensed under the MIT license. The full license text is available in the LICENSE file.\n',
    'author': 'Benedict Etzel',
    'author_email': 'benedict@hearthsim.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dj-stripe/dj-paypal',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import re
from datetime import datetime

from dateutil.parser import parse
from django.conf import settings
from django.db import models, transaction
from django.utils.timezone import now
from paypalrestsdk import payments as paypal_models

from .. import enums
from ..exceptions import AgreementAlreadyExecuted, PaypalApiError
from ..fields import CurrencyAmountField, JSONField
from ..settings import PAYPAL_LIVE_MODE
from .base import PaypalObject


def get_frequency_delta(frequency, frequency_interval):
	from dateutil.relativedelta import relativedelta

	frequency_kw = {
		enums.PaymentDefinitionFrequency.DAY: "days",
		enums.PaymentDefinitionFrequency.WEEK: "weeks",
		enums.PaymentDefinitionFrequency.MONTH: "months",
		enums.PaymentDefinitionFrequency.YEAR: "years",
	}[frequency.upper()]

	return relativedelta(**{frequency_kw: frequency_interval})


class BillingPlan(PaypalObject):
	name = models.CharField(max_length=128)
	description = models.CharField(max_length=127)
	type = models.CharField(max_length=20, choices=enums.BillingPlanType.choices)
	state = models.CharField(max_length=20, choices=enums.BillingPlanState.choices)
	create_time = models.DateTimeField(db_index=True, editable=False)
	update_time = models.DateTimeField(db_index=True, editable=False)
	merchant_preferences = JSONField()

	payment_definitions = models.ManyToManyField("PaymentDefinition")

	paypal_model = paypal_models.BillingPlan

	@classmethod
	def clean_api_data(cls, data):
		id, cleaned_data, m2ms = super().clean_api_data(data)

		pds = cleaned_data.pop("payment_definitions")
		# Sync payment definitions but do not fetch them (we have them in full)
		m2ms["payment_definitions"] = PaymentDefinition.objects.sync_data(pds, fetch=False)

		return id, cleaned_data, m2ms

	@classmethod
	def create(cls, data, activate=False):
		obj = cls.paypal_model(data)
		obj.create()
		if not obj:
			raise PaypalApiError("Could not create plan: %r" % (obj.error))

		instance, created = cls.get_or_update_from_api_data(obj, always_sync=True)
		if activate:
			instance.activate()

		return instance

	@property
	def regular_payment_definition(self):
		return self.payment_definitions.get(type=enums.PaymentDefinitionType.REGULAR)

	@property
	def human_readable_price(self):
		pd = self.regular_payment_definition
		if pd:
			return pd.human_readable_price
		return ""

	def activate(self):
		"""
		Activate an plan in a CREATED state.
		"""
		obj = self.find_paypal_object()
		if obj.state == enums.BillingPlanState.CREATED:
			success = obj.activate()
			if not success:
				raise PaypalApiError("Failed to activate plan: %r" % (obj.error))
		# Resync the updated data to the database
		self.get_or_update_from_api_data(obj, always_sync=True)
		return obj

	def create_agreement(
		self, user, start_date,
		payment_method="paypal", override_merchant_preferences=None
	):
		billing_agreement = paypal_models.BillingAgreement({
			"name": self.name,
			"description": self.description,
			"plan": {"id": self.id},
			"payer": {"payment_method": payment_method},
			"start_date": start_date.replace(microsecond=0).isoformat(),
		})

		if override_merchant_preferences:
			billing_agreement["override_merchant_preferences"] = override_merchant_preferences.copy()

		if not billing_agreement.create():
			raise PaypalApiError("Error creating Billing Agreement: %r" % (billing_agreement.error))

		return PreparedBillingAgreement.create_from_data(billing_agreement, user)


class PreparedBillingAgreement(models.Model):
	"""
	A class that stores a billing agreement execution token,
	and ties it to the user who requested it.
	"""
	id = models.CharField(
		max_length=128, primary_key=True, editable=False, serialize=True,
		help_text="Same as the BillingAgreement token"
	)
	livemode = models.BooleanField()
	user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
	data = JSONField()
	executed_agreement = models.ForeignKey(
		"BillingAgreement", on_delete=models.SET_NULL, null=True, blank=True,
		related_name="prepared_agreements"
	)
	executed_at = models.DateTimeField(null=True, blank=True)
	canceled_at = models.DateTimeField(null=True, blank=True)
	created = models.DateTimeField(auto_now_add=True, db_index=True)
	updated = models.DateTimeField(auto_now=True, db_index=True)

	@staticmethod
	def _extract_token(data):
		# Paypal does not provide a clear field for the token.
		# In order to offer a way to securely tie the user to a token, we
		# use it as a primary key in this model but we have to extract it
		# from the HATEOAS urls first.
		token_match = re.compile(r"/billing-agreements/([^/]+)/agreement-execute")
		for link in data.get("links", []):
			sre = token_match.search(link.get("href", ""))
			if sre:
				return sre.groups(0)[0]
		raise ValueError("Could not find token in billing agreement data")

	@classmethod
	def create_from_data(cls, data, user):
		data = data.to_dict()
		return cls.objects.create(
			id=cls._extract_token(data), livemode=PAYPAL_LIVE_MODE, user=user, data=data
		)

	def __str__(self):
		return "<{}: {}>".format(self.__class__.__name__, self.id)

	@property
	def approval_url(self):
		for link in self.data.get("links", []):
			if link["rel"] == "approval_url":
				return link["href"]
		return ""

	@property
	def plan_id(self):
		return self.data.get("plan", {}).get("id", None)

	def cancel(self):
		if self.executed_at:
			raise AgreementAlreadyExecuted("Agreement has already been executed")
		self.canceled_at = now()
		self.save()

	def execute(self):
		"""
		Execute the PreparedBillingAgreement by creating and executing a
		matching BillingAgreement.
		"""
		# Save the execution time first.
		# If execute() fails, executed_at will be set, with no executed_agreement set.
		self.executed_at = now()
		self.save()

		with transaction.atomic():
			ret = BillingAgreement.execute(self.id)
			ret.user = self.user
			ret.plan_model_id = self.plan_id
			ret.save()
			self.executed_agreement = ret
			self.save()

		return ret


class BillingAgreement(PaypalObject):
	name = models.CharField(max_length=128, blank=True)
	state = models.CharField(
		max_length=128, editable=False, choices=enums.BillingAgreementState.choices
	)
	description = models.CharField(max_length=128)
	start_date = models.DateTimeField(db_index=True)
	agreement_details = JSONField()
	payer = JSONField()
	shipping_address = JSONField(null=True)
	override_merchant_preferences = JSONField(default=dict, blank=True)
	override_charge_mode = JSONField(default=dict, blank=True)
	plan = JSONField()
	merchant = JSONField(null=True, blank=True)

	payer_model = models.ForeignKey(
		"Payer", on_delete=models.SET_NULL, null=True, blank=True,
	)
	plan_model = models.ForeignKey(
		BillingPlan, on_delete=models.SET_NULL, null=True, blank=True,
	)
	user = models.ForeignKey(
		settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, null=True, blank=True
	)
	end_of_period = models.DateTimeField(db_index=True)

	paypal_model = paypal_models.BillingAgreement
	dashboard_url_template = (
		"{webscr}?cmd=_profile-recurring-payments&encrypted_profile_id={id}"
	)

	@classmethod
	def clean_api_data(cls, data):
		id, cleaned_data, m2ms = super().clean_api_data(data)

		# Fix inconsistent US/UK spelling
		if "state" in cleaned_data and cleaned_data["state"].lower() == "canceled":
			cleaned_data["state"] = enums.BillingAgreementState.Cancelled

		return id, cleaned_data, m2ms

	@classmethod
	def execute(cls, token):
		if not token:
			raise ValueError("Invalid token argument")

		ba = cls.paypal_model.execute(token)
		if ba.error:
			raise PaypalApiError(str(ba.error))  # , ba.error)

		obj, created = cls.get_or_update_from_api_data(ba, always_sync=True)
		return obj

	def save(self, **kwargs):
		from .payer import Payer

		# On save, get the payer_info object and do a best effort attempt at
		# saving a Payer model and relation into the db.
		payer_info = self.payer.get("payer_info", {})
		if payer_info and "payer_id" in payer_info:
			# Copy the payer_info dict before mutating it
			payer_info = payer_info.copy()
			payer_id = payer_info.pop("payer_id")
			payer_info["user"] = self.user
			payer_info["livemode"] = self.livemode
			self.payer_model, created = Payer.objects.update_or_create(
				id=payer_id, defaults=payer_info
			)

		self.end_of_period = self.calculate_end_of_period()
		return super().save(**kwargs)

	def cancel(self, note, immediately=False):
		obj = self.find_paypal_object()
		obj.cancel({"note": note})
		# Sync updated object back to to the database
		obj, created = self.get_or_update_from_api_data(obj, always_sync=True)
		if immediately:
			obj.end_of_period = datetime.now()
			obj.save()
		return obj

	def suspend(self, note):
		obj = self.find_paypal_object()
		obj.suspend({"note": note})
		# Sync updated object back to to the database
		obj, created = self.get_or_update_from_api_data(obj, always_sync=True)
		return obj

	@property
	def last_payment_date(self):
		date = self.agreement_details.get("last_payment_date", "")
		if date:
			return parse(date)

	def calculate_end_of_period(self):
		# The next payment date is not reliably set.
		# When a subscription is cancelled, we do not have access to it anymore...
		# So instead, keep the end_of_period attribute up to date.
		last_payment_date = self.last_payment_date
		if not last_payment_date:
			return parse("1970-01-01T00:00:00Z")

		rpd = next(filter(
			lambda pd: pd["type"] == enums.PaymentDefinitionType.REGULAR,
			self.plan["payment_definitions"]
		))

		delta = get_frequency_delta(rpd["frequency"], int(rpd["frequency_interval"]))

		return last_payment_date + delta


class PaymentDefinition(PaypalObject):
	name = models.CharField(max_length=128)
	type = models.CharField(max_length=20, choices=enums.PaymentDefinitionType.choices)
	frequency_interval = models.PositiveSmallIntegerField()
	frequency = models.CharField(
		max_length=20, choices=enums.PaymentDefinitionFrequency.choices
	)
	cycles = models.PositiveSmallIntegerField()
	amount = CurrencyAmountField()

	charge_models = models.ManyToManyField("ChargeModel")

	@classmethod
	def clean_api_data(cls, data):
		id, cleaned_data, m2ms = super().clean_api_data(data)

		cleaned_data["frequency"] = cleaned_data["frequency"].upper()
		if "charge_models" in cleaned_data:
			charge_models = cleaned_data.pop("charge_models")
			# Sync payment definitions but do not fetch them (we have them in full)
			m2ms["charge_models"] = ChargeModel.objects.sync_data(charge_models, fetch=False)

		return id, cleaned_data, m2ms

	@property
	def human_readable_price(self):
		from ..utils import get_friendly_currency_amount

		amount = get_friendly_currency_amount(self.amount["value"], self.amount["currency"])
		interval_count = self.frequency_interval

		if interval_count == 1:
			interval = self.frequency.lower()
			template = "{amount}/{interval}"
		else:
			interval = {
				enums.PaymentDefinitionFrequency.DAY: "days",
				enums.PaymentDefinitionFrequency.WEEK: "weeks",
				enums.PaymentDefinitionFrequency.MONTH: "months",
				enums.PaymentDefinitionFrequency.YEAR: "years",
			}[self.frequency]
			template = "{amount} every {interval_count} {interval}"

		return template.format(amount=amount, interval=interval, interval_count=interval_count)

	@property
	def frequency_delta(self):
		"""
		Helper property that returns a dateutil.relativedelta.relativedelta
		object which represents a full billing period.

		https://dateutil.readthedocs.io/en/stable/relativedelta.html
		"""

		return get_frequency_delta(self.frequency, self.frequency_interval)


class ChargeModel(PaypalObject):
	type = models.CharField(max_length=20, choices=enums.ChargeModelType.choices)
	amount = CurrencyAmountField()



# Generated by Django 2.1.2 on 2018-11-18 15:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djpaypal', '0006_auto_20181001_1147'),
    ]

    operations = [
        migrations.AddField(
            model_name='sale',
            name='invoice_number',
            field=models.CharField(blank=True, help_text='Your own invoice or tracking ID number.', max_length=127),
        ),
    ]



# Generated by Django 2.1.2 on 2019-11-10 15:20

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djpaypal', '0009_shipping_address_null'),
    ]

    operations = [
        migrations.AddField(
            model_name='sale',
            name='payment_hold_reasons',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, editable=False, null=True),
        ),
        migrations.AddField(
            model_name='sale',
            name='payment_hold_status',
            field=models.CharField(choices=[('HELD', 'Held'), ('RELEASED', 'Released')], editable=False, max_length=8, null=True),
        ),
    ]



# Generated by Django 2.1.2 on 2019-11-10 15:20

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djpaypal', '0008_refund_refund_to_payer'),
    ]

    operations = [
        migrations.AlterField(
            model_name='billingagreement',
            name='shipping_address',
            field=django.contrib.postgres.fields.jsonb.JSONField(null=True),
        ),
    ]
